/************************************************************************
 *   IRC - Internet Relay Chat, m_core/m_topic.c
 *   This file is copyright (C) 2001 Andrew Suffield
 *                                    <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "ircd.h"
#include "common.h"
#include "numeric.h"
#include "irc_string.h"
#include "send.h"
#include "umodes.h"
#include "msg.h"
#include "handlers.h"
#include "channel.h"
#include "struct.h"
#include "hash.h"
#include "burst.h"
#include "cmodes.h"

#include <stdlib.h>
#include <string.h>

int m_sjoin(struct Client *, struct Client *, int, char **);

struct Message sjoin_msgtab = {
  "SJOIN", 0, 0, 0, MFLG_SLOW, 0,
  {m_unregistered, m_ignore, m_sjoin},
  NULL, NULL
};

static char modebuf[MODEBUFLEN], modebuf2[MODEBUFLEN];
static char parabuf[MODEBUFLEN], parabuf2[MODEBUFLEN];

static void
sjoin_sendit(struct Client *cptr,
	     struct Client *sptr, struct Channel *chptr, char *from)
{
  sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", from,
			 chptr->chname, modebuf, parabuf);
}

/*
 * m_sjoin
 * parv[0] - sender
 * parv[1] - TS
 * parv[2] - channel
 * parv[3] - modes + n arguments (key and/or limit)
 * parv[4+n] - flags+nick list (all in one parameter)

 * 
 * process a SJOIN, taking the TS's into account to either ignore the
 * incoming modes or undo the existing ones or merge them, and JOIN
 * all the specified users while sending JOIN/MODEs to non-TS servers
 * and to clients
 */


int
m_sjoin(struct Client *cptr, struct Client *sptr, int parc, char *parv[])
{
  struct Channel *chptr;
  struct Client *acptr;
  char *mbufw = modebuf, *mbuf2w = modebuf2;
  char *pbufw = parabuf, *pbuf2w = parabuf2;
  int len = 0;

  if (IsClient(sptr) || parc < 5)
    return -1;
  if (!IsChannelName(parv[2]))
    return -1;

  if (!check_channel_name(parv[2]))
    {
      sendto_ops_flag(UMODE_DEBUG, "WTF: invalid channel name %s from %s",
		      parv[2], sptr->name);
      return -1;
    }

  /* Under the new regime of pre-exchanging channel timestamps,
   * we will only get an SJOIN to add modes. There is no need to
   * do all that clever hackery to clear old ones out, that is done
   * completely and cleanly during the burst.
   *  -- asuffield
   */

  chptr = get_channel(sptr, parv[2], CREATE);

  chptr->channelts = atol(parv[1]);

  /* I am changing parv[3] to be merely mode changes, not a list of all the
   * new modes. When an SJOIN is sent, then all modes will be sent in it, not
   * just those which have changed (send_channel_modes is responsible for this)
   *  -- asuffield
   */

  {
    char *s, c;
    int whatt = MODE_QUERY;
    char plus = '=';

    /* Parse the 'global' channel modes. Build up a list of positive changes
     * and tag the mode as changed. Afterwards, go through the modes and
     * issue negative changes on the remaining untouched ones
     */
    s = parv[3];
    parc -= 4;
    parv += 4;
    while (*s)
      {
	switch (c = *s++)
	  {
	  case '+':
	    whatt = MODE_ADD;
	    plus = '+';
	    continue;

	  case '-':
	    whatt = MODE_DEL;
	    plus = '-';
	    continue;

	  case '=':
	    whatt = MODE_QUERY;
	    plus = '=';
	    continue;

	  default:
	    if (whatt == MODE_QUERY)
	      continue;
	    parse_mode_char(sptr, chptr, c, plus, whatt,
			    &mbufw, &mbuf2w, &pbufw, &pbuf2w,
			    &len, &parc, &parv);
	  }
      }
  }
  if (!parc)
    {
      sendto_ops_flag(UMODE_DEBUG,
		      "WTF: Got SJOIN with no parameters left after mode parsing on %s from %s",
		      chptr->chname, sptr->name);
      return -1;
    }

  /* OK, we've done channel flag things. Send out those changes to my clients
   * now, then we'll deal with the new user flags
   *
   * They'll get sent to the other servers when I pass out the SJOINs
   */
  *mbuf2w = *pbufw = *pbuf2w = '\0';
  strncpy_irc(mbufw, modebuf2, MODEBUFLEN - strlen(modebuf));
  collapse_signs(modebuf);

  if (*modebuf)
    {
      sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s %s",
			     IsServer(sptr) ? me.name : sptr->name,
			     chptr->chname, modebuf, parabuf, parabuf2);
    }

  {
    char sjoinbuf[BUFSIZE];
    register char *s, *s0;
    int pargs = 0, people = 0;	/* people is used to filter out SJOINs with nothing in them */
    char *sjbuf = sjoinbuf;
    char *p;			/* strtoken() uses this */

    /* OK, now build up the SJOIN while parsing the new user list */
    channel_modes(sptr, modebuf, parabuf, chptr, 0);

    sjbuf +=
      ircsprintf(sjbuf, ":%s SJOIN %lu %s %s %s :", sptr->name,
		 chptr->channelts, chptr->chname, modebuf, parabuf);

    mbufw = modebuf;
    parabuf[0] = '\0';
    *mbufw++ = '+';

    for (s = s0 = strtoken(&p, parv[0], " "); s;
	 s = s0 = strtoken(&p, (char *) NULL, " "))
      {
	/* Grab the channel user flags from the name prefix here */
	int fl = 0;
	for (;;)
	  {
	    switch (*s)
	      {
	      case '@':
		fl |= CUFLAG_CHANOP;
		s++;
		continue;
	      case '+':
		fl |= CUFLAG_VOICE;
		s++;
		continue;
	      default:
		break;
	      }
	    break;
	  }
	/* OK, can we find this guy? */
	if (!(acptr = find_chasing(sptr, s, NULL)))
	  continue;
	/* Are they coming from that way? If not, the SJOIN is invalid,
	 * that nick is being collided, the peer just sent this SJOIN
	 * before the KILL reached them
	 */
	if (acptr->from != cptr)
	  continue;
	people++;
	/* For safety, check this user isn't already joined, then add
	 * them and let my local clients know */
	if (!IsMember(acptr, chptr))
	  {
	    add_user_to_channel(chptr, acptr, fl);
	    sendto_channel_butserv(chptr, acptr, ":%s JOIN :%s", s, chptr->chname);	/* Note: s0 has the prefix, s does not */
	  }
	/* Add the user to the SJOIN I'm sending out, if I got this far
	 * then it's valid
	 */
	strcpy(sjbuf, s0);	/* Note: s0 has the prefix, s does not */
	sjbuf += strlen(sjbuf);
	*sjbuf++ = ' ';
	/* Let my local clients know about the channel user flags,
	 * they don't understand prefixes in JOIN
	 */
	if (fl & CUFLAG_CHANOP)
	  {
	    *mbufw++ = 'o';
	    strcat(parabuf, s);
	    strcat(parabuf, " ");
	    pargs++;
	    if (pargs >= 4)
	      {
		*mbufw = '\0';
		sjoin_sendit(&me, sptr, chptr, sptr->name);
		mbufw = modebuf;
		*mbufw++ = '+';
		parabuf[0] = '\0';
		pargs = 0;
	      }
	  }
	if (fl & CUFLAG_VOICE)
	  {
	    *mbufw++ = 'v';
	    strcat(parabuf, s);
	    strcat(parabuf, " ");
	    pargs++;
	    if (pargs >= 4)
	      {
		*mbufw = '\0';
		sjoin_sendit(&me, sptr, chptr, sptr->name);
		mbufw = modebuf;
		*mbufw++ = '+';
		parabuf[0] = '\0';
		pargs = 0;
	      }
	  }
      }

    *mbufw = '\0';
    if (pargs)
      sjoin_sendit(&me, sptr, chptr, sptr->name);
    if (people)
      {
	if (sjbuf[-1] == ' ')
	  sjbuf[-1] = '\0';
	else
	  *sjbuf = '\0';
	sendto_match_servs(chptr, cptr, "%s", sjoinbuf);
      }
  }

  return 0;
}
